/*
mjpegtools_yuv_to_v4l.c
Copyright (c) Jan Panteltje 2008-always
This software is distributed under the GNU public license version 2
See also the file 'COPYING'.
*/


#define VERSION "0.2"
#define READSIZE 65535

#define _GNU_SOURCE
#include <stdio.h>

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <sys/wait.h>
#include <linux/videodev.h>


int debug_flag;
char *buffer;
char *out_buffer;
int fmt = VIDEO_PALETTE_RGB24;
int read_img = 0;


char *strsave(char *s) /*save char array s somewhere*/
{
char *p;
 
p = malloc( strlen(s) +  1);
if(p)
	{
	strcpy(p, s);

	return p;
	}

return 0;
}


int start_pipe (int dev, int width, int height)
{
struct video_capability vid_caps;
struct video_window vid_win;
struct video_picture vid_pic;

if(ioctl (dev, VIDIOCGCAP, &vid_caps) == -1)
	{
	perror ("mjpegtools_yuv_to_v4l: ioctl (VIDIOCGCAP)");

	return 1;
	}

if(ioctl (dev, VIDIOCGPICT, &vid_pic) == -1)
	{
	perror ("mjpegtools_yuv_to_v4l: ioctl VIDIOCGPICT");

	return 1;
	}

vid_pic.palette = VIDEO_PALETTE_YUV420P;
if(ioctl (dev, VIDIOCSPICT, &vid_pic) == -1)
	{
	perror("mjpegtools_yuv_to_v4l: ioctl VIDIOCSPICT");

	return 1;
	}

if(ioctl (dev, VIDIOCGWIN, &vid_win) == -1)
	{
	perror ("mjpegtools_yuv_to_v4l: ioctl VIDIOCGWIN");

	return 1;
	}

vid_win.width = width;
vid_win.height = height;
if(ioctl (dev, VIDIOCSWIN, &vid_win) == -1)
	{
	perror("mjpegtools_yuv_to_v4l: ioctl VIDIOCSWIN");

	return 1;
	}

return 0;
}


int main (int argc, char **argv)
{
int devout;
int width;
int height;
char *bptr;
int a, b;
size_t sb;
char *running;
char *token;
int frame_rate;
char interlace_mode;
double aspect_ratio = 0.0;
int length, ylength, ulength, vlength;


debug_flag = 0;
	

fprintf(stderr, "mjpegtools_yuv_to_v4l-%s (c) Jan Panteltje 2008-always.\n", VERSION);

if (argc != 2)
	{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "mjpegtools_yuv_to_v4l output_device\n");
	fprintf(stderr, "input from stdin must be in mjpeg tools YUV format.\n");
	fprintf(stderr, "Example:\n");
	fprintf(stderr, "modprobe vloopback (from flashcam-1.1.tgz from http://www.swift-tools.net/Flashcam/)\n");
	fprintf(stderr, "check with dmesg what is input and output.\n");
	fprintf(stderr, "mcamip -a IPADDRESS -p PORT -u USERNAME -w PASSWORD -y | mjpegtools_yuv_to_v4l /dev/video2\n");
	fprintf(stderr, "test with: xawtv -c /dev/video1 -geometry 640x480\n");
	fprintf(stderr, "\n");
	exit(1);
	}


devout = open (argv[1], O_RDWR);
if(devout < 0)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: failed to open output video device, aborting.");

	exit(1);
	}

/* read YUV header */

/*
From mjpegtools:
Description of the (new!, forever?) YUV4MPEG2 stream format:

STREAM consists of
 o one '\n' terminated STREAM-HEADER
 o unlimited number of FRAMEs

FRAME consists of
 o one '\n' terminated FRAME-HEADER
 o "length" octets of planar YCrCb 4:2:0 image data
   (if frame is interlaced, then the two fields are interleaved)

STREAM-HEADER consists of
 o string "YUV4MPEG2 "  (note the space after the '2')
 o unlimited number of ' ' separated TAGGED-FIELDs
 o '\n' line terminator

FRAME-HEADER consists of
 o string "FRAME "  (note the space after the 'E')
 o unlimited number of ' ' separated TAGGED-FIELDs
 o '\n' line terminator

TAGGED-FIELD consists of
 o single ascii character tag
 o VALUE (which does not contain whitespace)

VALUE consists of
 o integer (base 10 ascii representation)
 or o RATIO
 or o single ascii character
 or o generic ascii string

RATIO consists of
 o numerator (integer)
 o ':' (a colon)
 o denominator (integer)

The currently supported tags for the STREAM-HEADER:
	 W - [integer] frame width, pixels, should be > 0
 H - [integer] frame height, pixels, should be > 0
 I - [char] interlacing:  p - progressive (none)
                          t - top-field-first
                          b - bottom-field-first
                          ? - unknown
 F - [ratio] frame-rate, 0:0 == unknown
 A - [ratio] sample (pixel) aspect ratio, 0:0 == unknown
 X - [character string] 'metadata' (unparsed, but passed around)

The currently supported tags for the FRAME-HEADER:
 X - character string 'metadata' (unparsed, but passed around)
*/

/* read stream header to get width and height and interlace */

/*
YUV4MPEG2 W720 H576 F25:1 I? A0:0 XM2AR002
FRAME
data
*/

/* create a line buffer */
bptr = malloc(READSIZE);
if(! bptr)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: malloc bptr failed\n");

	exit(1);
	}

/*
From libc.info:
ssize_t getline (char **LINEPTR, size_t *N, FILE *STREAM)
This function is a GNU extension, but it is the recommended way to
read lines from a stream.  The alternative standard functions are 
unreliable.
*/

b = READSIZE;
a = getline(&bptr, &sb, stdin);
if(a == -1)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: EOF in read stream header, stop.\n");

	exit(0);
	}

if(strncmp(bptr, "YUV4MPEG2", 9) != 0)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: YUV4MPEG2 not found in stream header, aborting.\n");

	exit(1);
	}

/* parse arguments */
running = strsave(bptr);
if(! running)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: strsave(options) failed, aborting.\n");

	exit(1);
	}

while(1)
	{
	token = strsep (&running, " ");
	if(token == NULL) break;

	/* avoid empty string */
	if(token[0] == 0) continue;
			
	if(strncmp(token, "W", 1) == 0)
		{
		sscanf(token, "W%d", &width);
		}

	sscanf(token, "H%d", &height);
	sscanf(token, "F%d:%d", &frame_rate, &a);
	sscanf(token, "I%c", &interlace_mode);
	} /* end while parse tags */

free(running);

if(debug_flag)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: width=%d height=%d\n\
	frame_rate=%d a=%d\n\
	interlace_mode=%c\n\
	aspect_ratio=%.2f\n",
	width, height,\
	frame_rate, a,\
	interlace_mode,\
	aspect_ratio);
	}

/* calculate length */
ylength = width * height;
ulength = (width / 2) * (height / 2);
vlength = (width / 2) * (height / 2);
length =  ylength + ulength + vlength;

buffer = malloc(length);
if(! buffer)
	{
	fprintf(stderr, "mjpegtools_yuv_to_v4l: could not allocate space for buffer, aborting.\n");

	exit(1);
	}

start_pipe(devout, width, height);

fprintf(stderr, "mjpegtools_yuv_to_v4l: Starting video stream.\n");

while(1)
	{
	/* read frames in YUV format from stdin */
	/* read a frame */
	b = READSIZE;
	a = getline(&bptr, &sb, stdin);
	if(a == -1)
		{
		fprintf(stderr, "mjpegtools_yuv_to_v4l: unexpected EOF in frame header, aborting.\n");
		
		exit(1);
		}

	if(strncmp(bptr, "FRAME", 5) != 0)
		{
		fprintf(stderr, "mjpegtools_yuv_to_v4l: No FRAME found in frame, aborting.\n");

		exit(1);
		}

	/* 
	This is how it is written:
	Write Y', Cb, and Cr planes
	for (i = 0; i < 3; i++)
		{
		unsigned char *srctop = upper_field[i];
		unsigned char *srcbot = lower_field[i];
		// alternately write one line from each
		for (y = 0; y < height; y += 2)
			{
			if (y4m_write(fd, srctop, width)) return Y4M_ERR_SYSTEM;
			srctop += width;
			if (y4m_write(fd, srcbot, width)) return Y4M_ERR_SYSTEM;
			srcbot += width;
			}
		// for chroma, width/height are half as big
		if (i == 0)
			{
			width /= 2;
			height /= 2;
			}
		}
	*/

	/* read in data */
	a = fread(buffer, sizeof(char), length, stdin);
	if(a != length)
		{
		fprintf(stderr, "mjpegtools_yuv_to_v4l: EOF in frame data, stop.\n");

		/* this is not an error ? */
		exit(0);
		}

	a = write(devout, buffer, length);
		{
		if(a != length)
	 	   {
			fprintf(stderr, "mjpegtools_yuv_to_v4l: Error writing to pipe, aborting.\n");

			exit(1);
			}
		}
	}

exit(0);
} /* end function main */



